@@ -371,6 +371,7 @@ vmwgfx_download_from_hw(struct saa_driver *driver, PixmapPtr pixmap,
 	goto out_err;
     REGION_SUBTRACT(vsaa->pScreen, &spix->dirty_hw, &spix->dirty_hw, readback);
     REGION_UNINIT(vsaa->pScreen, &intersection);
+
     return TRUE;
  out_err:
     REGION_UNINIT(vsaa->pScreen, &intersection);
@@ -522,9 +523,6 @@ vmwgfx_destroy_pixmap(struct saa_driver *driver, PixmapPtr pixmap)
     vmwgfx_pixmap_remove_present(vpix);
     WSBMLISTDELINIT(&vpix->pixmap_list);
     WSBMLISTDELINIT(&vpix->sync_x_head);
-
-    if (vpix->hw_is_dri2_fronts)
-	LogMessage(X_ERROR, "Incorrect dri2 front count.\n");
 }
 
 
@@ -610,9 +608,9 @@ vmwgfx_pix_resize(PixmapPtr pixmap, unsigned int old_pitch,
     }
 
     if (vpix->hw) {
-	if (xa_surface_redefine(vpix->hw, draw->width, draw->height,
-				draw->depth, xa_type_argb,
-				xa_format_unknown, vpix->xa_flags, 1) != 0)
+      if (!vmwgfx_xa_surface_redefine(vpix, vpix->hw, draw->width,
+				      draw->height, draw->depth, xa_type_argb,
+				      xa_format_unknown, vpix->xa_flags, 1))
 	    return FALSE;
     }
 
@@ -694,7 +692,8 @@ vmwgfx_modify_pixmap_header (PixmapPtr pixmap, int w, int h, int depth,
 
     vmwgfx_pix_resize(pixmap, old_pitch, old_height, old_width);
     vmwgfx_pixmap_free_storage(vpix);
-    WSBMLISTADDTAIL(&vpix->pixmap_list, &vsaa->pixmaps);
+    if (WSBMLISTEMPTY(&vpix->pixmap_list))
+	WSBMLISTADDTAIL(&vpix->pixmap_list, &vsaa->pixmaps);
 
     return TRUE;
 
@@ -798,7 +797,8 @@ vmwgfx_prefer_gmr(struct vmwgfx_saa *vsaa, PixmapPtr pixmap)
 
 Bool
 vmwgfx_create_hw(struct vmwgfx_saa *vsaa,
-		 PixmapPtr pixmap)
+		 PixmapPtr pixmap,
+		 Bool shared)
 {
     struct vmwgfx_saa_pixmap *vpix = vmwgfx_saa_pixmap(pixmap);
     struct xa_surface *hw;
@@ -807,28 +807,34 @@ vmwgfx_create_hw(struct vmwgfx_saa *vsaa,
     if (!vsaa->xat)
 	return FALSE;
 
-    if (vpix->hw)
-	return TRUE;
+    if (!shared) {
+	if (vpix->hw)
+	    return TRUE;
 
-    new_flags = (vpix->xa_flags & ~vpix->staging_remove_flags) |
-	vpix->staging_add_flags | XA_FLAG_SHARED;
+	new_flags = (vpix->xa_flags & ~vpix->staging_remove_flags) |
+	    vpix->staging_add_flags | XA_FLAG_SHARED;
+
+	hw = xa_surface_create(vsaa->xat,
+			       pixmap->drawable.width,
+			       pixmap->drawable.height,
+			       0,
+			       xa_type_other,
+			       vpix->staging_format,
+			       new_flags);
+    } else {
+	new_flags = vpix->xa_flags;
+	hw = vpix->hw;
+    }
 
-    hw = xa_surface_create(vsaa->xat,
-			   pixmap->drawable.width,
-			   pixmap->drawable.height,
-			   0,
-			   xa_type_other,
-			   vpix->staging_format,
-			   new_flags);
     if (hw == NULL)
 	return FALSE;
 
     vpix->xa_flags = new_flags;
+    vpix->hw = hw;
 
     if (!vmwgfx_pixmap_add_damage(pixmap))
 	goto out_no_damage;
 
-    vpix->hw = hw;
     vpix->backing |= VMWGFX_PIX_SURFACE;
     vmwgfx_pixmap_free_storage(vpix);
 
@@ -841,6 +847,7 @@ vmwgfx_create_hw(struct vmwgfx_saa *vsaa,
     return TRUE;
 
 out_no_damage:
+    vpix->hw = NULL;
     xa_surface_destroy(hw);
     return FALSE;
 }
@@ -1158,6 +1165,8 @@ vmwgfx_composite_prepare(struct saa_driver *driver, CARD8 op,
      * and check whether XA can accelerate.
      */
 
+    if (!mask_pix)
+	mask_pict = NULL;
     xa_comp = vmwgfx_xa_setup_comp(vsaa->vcomp, op,
 				   src_pict, mask_pict, dst_pict);
     if (!xa_comp)
@@ -1314,7 +1323,7 @@ vmwgfx_dirty(struct saa_driver *driver, PixmapPtr pixmap,
      * just before we call the kms update function for the hw
      * surface.
      */
-    if (vsaa->only_hw_presents) {
+    if (vpix->scanout_hw) {
 	if (!hw && !vmwgfx_upload_to_hw(&vsaa->driver, pixmap, damage))
 	    return FALSE;
 
@@ -1404,7 +1413,8 @@ vmwgfx_saa_init(ScreenPtr pScreen, int drm_fd, struct xa_tracker *xat,
 		void (*present_flush)(ScreenPtr pScreen),
 		Bool direct_presents,
 		Bool only_hw_presents,
-		Bool rendercheck)
+		Bool rendercheck,
+		Bool has_screen_targets)
 {
     struct vmwgfx_saa *vsaa;
 
@@ -1415,6 +1425,7 @@ vmwgfx_saa_init(ScreenPtr pScreen, int drm_fd, struct xa_tracker *xat,
     if (xat == NULL) {
 	direct_presents = FALSE;
 	only_hw_presents = FALSE;
+	has_screen_targets = FALSE;
     }
 
     vsaa->pScreen = pScreen;
@@ -1429,6 +1440,7 @@ vmwgfx_saa_init(ScreenPtr pScreen, int drm_fd, struct xa_tracker *xat,
     vsaa->rendercheck = rendercheck;
     vsaa->is_master = TRUE;
     vsaa->known_prime_format = FALSE;
+    vsaa->has_screen_targets = has_screen_targets;
     WSBMINITLISTHEAD(&vsaa->sync_x_list);
     WSBMINITLISTHEAD(&vsaa->pixmaps);
 
@@ -1488,7 +1500,8 @@ vmwgfx_scanout_refresh(PixmapPtr pixmap)
  */
 
 uint32_t
-vmwgfx_scanout_ref(struct vmwgfx_screen_entry  *entry)
+vmwgfx_scanout_ref(struct vmwgfx_screen_entry  *entry,
+		   Bool scanout_equals_pixmap)
 {
     PixmapPtr pixmap = entry->pixmap;
     struct vmwgfx_saa *vsaa =
@@ -1499,12 +1512,16 @@ vmwgfx_scanout_ref(struct vmwgfx_screen_entry  *entry)
 	uint32_t handle, dummy;
 	unsigned int depth;
 
-	if (vsaa->only_hw_presents) {
+	vpix->scanout_hw = vsaa->only_hw_presents ||
+	    (vsaa->has_screen_targets && scanout_equals_pixmap);
+
+	if (vpix->scanout_hw) {
 	    /*
 	     * The KMS fb will be a HW surface. Create it, add damage
 	     * and get the handle.
 	     */
-	    if (!vmwgfx_hw_accel_validate(pixmap, 0, XA_FLAG_SCANOUT, 0, NULL))
+	    if (!vmwgfx_hw_accel_validate(pixmap, 0, XA_FLAG_SCANOUT |
+					  XA_FLAG_RENDER_TARGET, 0, NULL))
 		goto out_err;
 	    if (_xa_surface_handle(vpix->hw, &handle, &dummy) != 0)
 		goto out_err;
